/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Texture.h            |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 01-11-2006
 *
 *	bool CGLTexture::b_State() const renamed to bool CGLTexture::b_Status() const
 *	(want to have unified naming over the whole OpenGL framework)
 *
 */

#include <vector>
#include "../berLame_src/CallStack.h"
#include <windows.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/glext.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "Texture.h"

/*
 *								=== CGLTextureParams ===
 */

const float CGLTextureParams::m_f_compare_epsilon = 1e-3f;

/*
 *	CGLTextureParams::CGLTextureParams()
 *		- default constructor
 *		- fill itself with default texture parameters
 */
CGLTextureParams::CGLTextureParams()
	:m_n_minify_filter(GL_NEAREST_MIPMAP_LINEAR), m_n_magnify_filter(GL_LINEAR),
	m_f_min_lod(-1000), m_f_max_lod(1000), m_f_lod_bias(0), m_n_base_level(0),
	m_n_max_level(1000), m_f_priority(1), m_n_depth_mode(GL_LUMINANCE),
	m_n_compare_mode(GL_NONE), m_n_compare_func(GL_LEQUAL), m_b_generate_mipmap(false)
{
	__FuncGuard("CGLTextureParams::CGLTextureParams");

	m_n_wrap_mode[coord_S] = GL_REPEAT;
	m_n_wrap_mode[coord_T] = GL_REPEAT;
	m_n_wrap_mode[coord_R] = GL_REPEAT;
	m_p_border_color[0] = 0;
	m_p_border_color[1] = 0;
	m_p_border_color[2] = 0;
	m_p_border_color[3] = 0;
}

/*
 *	void CGLTextureParams::UpdateState(GLenum n_target = GL_TEXTURE_2D)
 *		- get texture parameters from OpenGL
 *		- note texture of type n_target must be bound and enabled
 *		- OpenGL error(s) may be generated, call glGetError() afterwards
 */
void CGLTextureParams::UpdateState(GLenum n_target)
{
	__FuncGuard("CGLTextureParams::UpdateState");

	glGetTexParameteriv(n_target, GL_TEXTURE_WRAP_S, (int*)&m_n_wrap_mode[coord_S]);
	glGetTexParameteriv(n_target, GL_TEXTURE_WRAP_T, (int*)&m_n_wrap_mode[coord_T]);
	glGetTexParameteriv(n_target, GL_TEXTURE_WRAP_R, (int*)&m_n_wrap_mode[coord_R]);
	glGetTexParameteriv(n_target, GL_TEXTURE_MIN_FILTER, (int*)&m_n_minify_filter);
	glGetTexParameteriv(n_target, GL_TEXTURE_MAG_FILTER, (int*)&m_n_magnify_filter);
	glGetTexParameterfv(n_target, GL_TEXTURE_BORDER_COLOR, m_p_border_color);
	glGetTexParameterfv(n_target, GL_TEXTURE_PRIORITY, &m_f_priority);
	glGetTexParameterfv(n_target, GL_TEXTURE_MIN_LOD, &m_f_min_lod);
	glGetTexParameterfv(n_target, GL_TEXTURE_MAX_LOD, &m_f_max_lod);
	glGetTexParameteriv(n_target, GL_TEXTURE_BASE_LEVEL, &m_n_base_level);
	glGetTexParameteriv(n_target, GL_TEXTURE_MAX_LEVEL, &m_n_max_level);
	glGetTexParameterfv(n_target, GL_TEXTURE_LOD_BIAS, &m_f_lod_bias);
	glGetTexParameteriv(n_target, GL_DEPTH_TEXTURE_MODE, (int*)&m_n_depth_mode);
	glGetTexParameteriv(n_target, GL_TEXTURE_COMPARE_MODE, (int*)&m_n_compare_mode);
	glGetTexParameteriv(n_target, GL_TEXTURE_COMPARE_FUNC, (int*)&m_n_compare_func);
	int n_temp;
	glGetTexParameteriv(n_target, GL_GENERATE_MIPMAP, &n_temp);
	m_b_generate_mipmap = (n_temp == GL_TRUE);
}

/*
 *	void CGLTextureParams::SetState(GLenum n_target = GL_TEXTURE_2D) const
 *		- set texture parameters to OpenGL
 *		- note texture of type n_target must be bound and enabled
 *		- OpenGL error(s) may be generated, call glGetError() afterwards
 */
void CGLTextureParams::SetState(GLenum n_target) const
{
	__FuncGuard("CGLTextureParams::SetState");

	glTexParameteri(n_target, GL_TEXTURE_WRAP_S, m_n_wrap_mode[coord_S]);
	glTexParameteri(n_target, GL_TEXTURE_WRAP_T, m_n_wrap_mode[coord_T]);
	glTexParameteri(n_target, GL_TEXTURE_WRAP_R, m_n_wrap_mode[coord_R]);
	glTexParameteri(n_target, GL_TEXTURE_MIN_FILTER, m_n_minify_filter);
	glTexParameteri(n_target, GL_TEXTURE_MAG_FILTER, m_n_magnify_filter);
	glTexParameterfv(n_target, GL_TEXTURE_BORDER_COLOR, m_p_border_color);
	glTexParameterf(n_target, GL_TEXTURE_PRIORITY, m_f_priority);
	glTexParameterf(n_target, GL_TEXTURE_MIN_LOD, m_f_min_lod);
	glTexParameterf(n_target, GL_TEXTURE_MAX_LOD, m_f_max_lod);
	glTexParameteri(n_target, GL_TEXTURE_BASE_LEVEL, m_n_base_level);
	glTexParameteri(n_target, GL_TEXTURE_MAX_LEVEL, m_n_max_level);
	glTexParameterf(n_target, GL_TEXTURE_LOD_BIAS, m_f_lod_bias);
	glTexParameteri(n_target, GL_DEPTH_TEXTURE_MODE, m_n_depth_mode);
	glTexParameteri(n_target, GL_TEXTURE_COMPARE_MODE, m_n_compare_mode);
	glTexParameteri(n_target, GL_TEXTURE_COMPARE_FUNC, m_n_compare_func);
	glTexParameteri(n_target, GL_GENERATE_MIPMAP, (m_b_generate_mipmap)? GL_TRUE : GL_FALSE);
}

/*
 *	void CGLTextureParams::SetState(GLenum n_target, const CGLTextureParams &r_state) const
 *		- set texture parameters different from corresponding parameters in r_state to OpenGL
 *		  (effect is the same as calling r_state.SetState() but equal values aren't set
 *		  so OpenGL call overhead is lower which should result in better performance)
 *		- note floating-point number comparison is done with epsilon = 1e-3
 *		- note texture of type n_target must be bound and enabled
 *		- OpenGL error(s) may be generated, call glGetError() afterwards
 */
void CGLTextureParams::SetState(GLenum n_target, const CGLTextureParams &r_state) const
{
	__FuncGuard("CGLTextureParams::SetState");

	if(m_n_wrap_mode[coord_S] != r_state.m_n_wrap_mode[coord_S])
		glTexParameteri(n_target, GL_TEXTURE_WRAP_S, m_n_wrap_mode[coord_S]);
	if(m_n_wrap_mode[coord_T] != r_state.m_n_wrap_mode[coord_T])
		glTexParameteri(n_target, GL_TEXTURE_WRAP_T, m_n_wrap_mode[coord_T]);
	if(m_n_wrap_mode[coord_R] != r_state.m_n_wrap_mode[coord_R])
		glTexParameteri(n_target, GL_TEXTURE_WRAP_R, m_n_wrap_mode[coord_R]);
	if(m_n_minify_filter != r_state.m_n_minify_filter)
		glTexParameteri(n_target, GL_TEXTURE_MIN_FILTER, m_n_minify_filter);
	if(m_n_magnify_filter != r_state.m_n_magnify_filter)
		glTexParameteri(n_target, GL_TEXTURE_MAG_FILTER, m_n_magnify_filter);
	if(fabs(m_p_border_color[0] - r_state.m_p_border_color[0]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[1] - r_state.m_p_border_color[1]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[2] - r_state.m_p_border_color[2]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[3] - r_state.m_p_border_color[3]) > m_f_compare_epsilon)
		glTexParameterfv(n_target, GL_TEXTURE_BORDER_COLOR, m_p_border_color);
	if(fabs(m_f_priority - r_state.m_f_priority) > m_f_compare_epsilon)
		glTexParameterf(n_target, GL_TEXTURE_PRIORITY, m_f_priority);
	if(fabs(m_f_min_lod - r_state.m_f_min_lod) > m_f_compare_epsilon)
		glTexParameterf(n_target, GL_TEXTURE_MIN_LOD, m_f_min_lod);
	if(fabs(m_f_max_lod - r_state.m_f_max_lod) > m_f_compare_epsilon)
		glTexParameterf(n_target, GL_TEXTURE_MAX_LOD, m_f_max_lod);
	if(m_n_max_level != r_state.m_n_max_level)
		glTexParameteri(n_target, GL_TEXTURE_MAX_LEVEL, m_n_max_level);
	if(m_n_base_level != r_state.m_n_base_level)
		glTexParameteri(n_target, GL_TEXTURE_BASE_LEVEL, m_n_base_level);
	if(fabs(m_f_lod_bias - r_state.m_f_lod_bias) > m_f_compare_epsilon)
		glTexParameterf(n_target, GL_TEXTURE_LOD_BIAS, m_f_lod_bias);
	if(m_n_depth_mode != r_state.m_n_depth_mode)
		glTexParameteri(n_target, GL_DEPTH_TEXTURE_MODE, m_n_depth_mode);
	if(m_n_compare_mode != r_state.m_n_compare_mode)
		glTexParameteri(n_target, GL_TEXTURE_COMPARE_MODE, m_n_compare_mode);
	if(m_n_compare_func != r_state.m_n_compare_func)
		glTexParameteri(n_target, GL_TEXTURE_COMPARE_FUNC, m_n_compare_func);
	if(m_b_generate_mipmap != r_state.m_b_generate_mipmap)
		glTexParameteri(n_target, GL_GENERATE_MIPMAP, (m_b_generate_mipmap)? GL_TRUE : GL_FALSE);
}

/*
 *	bool CGLTextureParams::operator ==(const CGLTextureParams &r_state) const
 *		- compare two texture parameters objects, return true if equal, otherwise false
 *		- does no changes to OpenGL state, needs no particular state nor generates any errors
 *		- note floating-point number comparison is done with epsilon = 1e-3
 */
bool CGLTextureParams::operator ==(const CGLTextureParams &r_state) const
{
	__FuncGuard("CGLTextureParams::operator ==");

	if(m_n_wrap_mode[coord_S] != r_state.m_n_wrap_mode[coord_S] ||
	   m_n_wrap_mode[coord_T] != r_state.m_n_wrap_mode[coord_T] ||
	   m_n_wrap_mode[coord_R] != r_state.m_n_wrap_mode[coord_R] ||
	   m_n_minify_filter != r_state.m_n_minify_filter ||
	   m_n_magnify_filter != r_state.m_n_magnify_filter ||
	   m_n_max_level != r_state.m_n_max_level ||
	   m_n_base_level != r_state.m_n_base_level ||
	   m_n_depth_mode != r_state.m_n_depth_mode ||
	   m_n_compare_mode != r_state.m_n_compare_mode ||
	   m_n_compare_func != r_state.m_n_compare_func ||
	   m_b_generate_mipmap != r_state.m_b_generate_mipmap)
		return false;
	if(fabs(m_f_lod_bias - r_state.m_f_lod_bias) > m_f_compare_epsilon ||
	   fabs(m_f_priority - r_state.m_f_priority) > m_f_compare_epsilon ||
	   fabs(m_f_min_lod - r_state.m_f_min_lod) > m_f_compare_epsilon ||
	   fabs(m_f_max_lod - r_state.m_f_max_lod) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[0] - r_state.m_p_border_color[0]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[1] - r_state.m_p_border_color[1]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[2] - r_state.m_p_border_color[2]) > m_f_compare_epsilon ||
	   fabs(m_p_border_color[3] - r_state.m_p_border_color[3]) > m_f_compare_epsilon)
		return false;
	return true;
}

/*
 *								=== ~CGLTextureParams ===
 */

/*
 *								=== CGLTexture ===
 */

/*
 *	CGLTexture::CGLTexture()
 *		- default constructor
 *		- creates no texture object
 */
CGLTexture::CGLTexture()
	:m_n_id(0)
{
	__FuncGuard("CGLTexture::CGLTexture");
}

/*
 *	CGLTexture::~CGLTexture()
 *		- default destructor
 *		- deletes OpenGL texture object in case it was created
 */
CGLTexture::~CGLTexture()
{
	__FuncGuard("CGLTexture::~CGLTexture");

	if(m_n_id)
		glDeleteTextures(1, &m_n_id);
}

/*
 *	bool CGLTexture::b_Status() const
 *		- returns true in case texture was successfuly created and can be used for texturing
 */
bool CGLTexture::b_Status() const
{
	__FuncGuard("CGLTexture::b_Status");

	return m_n_id > 0 && glIsTexture(m_n_id);
}

/*
 *	void CGLTexture::Bind(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 */
void CGLTexture::Bind(CGLState *p_state) const
{
	__FuncGuard("CGLTexture::Bind");

	p_state->BindTexture(m_n_target, m_n_id);
}

/*
 *	void CGLTexture::Bind_Enable(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 *		- enables application of texture target (i.e. GL_TEXTURE_2D or similar)
 */
void CGLTexture::Bind_Enable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture::Bind_Enable");

	p_state->BindTexture(m_n_target, m_n_id);
	p_state->EnableTexture(m_n_target);
}

/*
 *	void CGLTexture::Disable(CGLState *p_state) const
 *		- disables application of texture target (i.e. GL_TEXTURE_2D or similar)
 */
void CGLTexture::Disable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture::Disable");

	p_state->DisableTexture(m_n_target);
}

/*
 *	CGLTextureParams &CGLTexture::r_Parameters(CGLState *p_state)
 *		- enables texture parameters to be changed
 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
 */
CGLTextureParams &CGLTexture::r_Parameters(CGLState *p_state)
{
	__FuncGuard("CGLTexture::r_Parameters");

	p_state->BindTexture(m_n_target, m_n_id);
	p_state->EnableTexture(m_n_target);
	return m_tex_parameters;
}

/*
 *								=== ~CGLTexture ===
 */

/*
 *								=== CGLTexture_1D ===
 */

CGLTexture_1D::CGLTexture_1D(CGLState *p_state, int n_width, GLenum n_internal_format,
	bool b_create_mipmaps, int n_border_width, GLenum n_target,
	GLenum n_format, GLenum n_data_type, const void *p_src_data)
	:m_n_width(n_width)
{
	__FuncGuard("CGLTexture_1D::CGLTexture_1D");

	m_n_internal_format = n_internal_format;
	m_n_border_width = n_border_width;
	m_n_target = n_target;

	glGenTextures(1, &m_n_id);
	if(m_n_id > 0) {
		p_state->EnableTexture1D();
		p_state->DisableTexture2D();
		p_state->DisableTexture3D();
		p_state->DisableTextureCubeMap();

		p_state->BindTexture1D(m_n_id);
		
		if(b_create_mipmaps)
			glTexParameteri(m_n_target, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage1D(m_n_target, 0, m_n_internal_format, m_n_width,
			m_n_border_width, n_format, n_data_type, (p_src_data)? p_src_data : NULL);

		if(!b_create_mipmaps)
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		// todo - see if this is necessary

		m_tex_parameters.UpdateState(m_n_target);

		if(glGetError() != GL_NO_ERROR) {
			glDeleteTextures(1, &m_n_id);
			m_n_id = 0;
		}
	}
}

/*
 *	void CGLTexture_1D::Bind(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 */
void CGLTexture_1D::Bind(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_1D::Bind");

	p_state->BindTexture1D(m_n_id);
}

/*
 *	void CGLTexture_1D::Bind_Enable(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 *		- enables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_1D::Bind_Enable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_1D::Bind_Enable");

	p_state->EnableTexture1D();
	p_state->BindTexture1D(m_n_id);
}

/*
 *	void CGLTexture_1D::Disable(CGLState *p_state) const
 *		- disables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_1D::Disable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_1D::Disable");

	p_state->DisableTexture1D();
}

/*
 *	CGLTextureParams &CGLTexture_1D::r_Parameters(CGLState *p_state)
 *		- enables texture parameters to be changed
 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
 */
CGLTextureParams &CGLTexture_1D::r_Parameters(CGLState *p_state)
{
	__FuncGuard("CGLTexture_1D::r_Parameters");

	p_state->EnableTexture1D();
	p_state->BindTexture1D(m_n_id);
	return m_tex_parameters;
}

/*
 *								=== ~CGLTexture_1D ===
 */

/*
 *								=== CGLTexture_2D ===
 */

CGLTexture_2D::CGLTexture_2D(CGLState *p_state, int n_width, int n_height,
	GLenum n_internal_format, bool b_create_mipmaps, int n_border_width, GLenum n_target,
	GLenum n_format, GLenum n_data_type, const void *p_src_data)
	:m_n_width(n_width), m_n_height(n_height)
{
	__FuncGuard("CGLTexture_2D::CGLTexture_2D");

	m_n_internal_format = n_internal_format;
	m_n_border_width = n_border_width;
	m_n_target = n_target;

	glGenTextures(1, &m_n_id);
	if(m_n_id > 0) {
		p_state->DisableTexture1D();
		p_state->EnableTexture2D();
		p_state->DisableTexture3D();
		p_state->DisableTextureCubeMap();

		p_state->BindTexture2D(m_n_id);
		
		if(b_create_mipmaps)
			glTexParameteri(m_n_target, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage2D(m_n_target, 0, m_n_internal_format, m_n_width, m_n_height,
			m_n_border_width, n_format, n_data_type, (p_src_data)? p_src_data : NULL);

		if(!b_create_mipmaps) {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		} else {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
		// t_odo - see if this is necessary - seems to be ok

		m_tex_parameters.UpdateState(m_n_target);

		if(glGetError() != GL_NO_ERROR) {
			glDeleteTextures(1, &m_n_id);
			m_n_id = 0;
		}
	}
}

/*
 *	void CGLTexture_2D::Bind(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 */
void CGLTexture_2D::Bind(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_2D::Bind");

	p_state->BindTexture2D(m_n_id);
}

/*
 *	void CGLTexture_2D::Bind_Enable(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 *		- enables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_2D::Bind_Enable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_2D::Bind_Enable");

	p_state->EnableTexture2D();
	p_state->BindTexture2D(m_n_id);
}

/*
 *	void CGLTexture_2D::Disable(CGLState *p_state) const
 *		- disables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_2D::Disable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_2D::Disable");

	p_state->DisableTexture2D();
}

/*
 *	CGLTextureParams &CGLTexture_2D::r_Parameters(CGLState *p_state)
 *		- enables texture parameters to be changed
 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
 */
CGLTextureParams &CGLTexture_2D::r_Parameters(CGLState *p_state)
{
	__FuncGuard("CGLTexture_2D::r_Parameters");

	p_state->EnableTexture2D();
	p_state->BindTexture2D(m_n_id);
	return m_tex_parameters;
}

/*
 *								=== ~CGLTexture_2D ===
 */

/*
 *								=== CGLTexture_3D ===
 */

CGLTexture_3D::CGLTexture_3D(CGLState *p_state, int n_width, int n_height, int n_depth,
	GLenum n_internal_format, bool b_create_mipmaps, int n_border_width,
	GLenum n_target, GLenum n_format, GLenum n_data_type, const void *p_src_data)
	:m_n_width(n_width), m_n_height(n_height), m_n_depth(n_depth)
{
	__FuncGuard("CGLTexture_3D::CGLTexture_3D");

	m_n_internal_format = n_internal_format;
	m_n_border_width = n_border_width;
	m_n_target = n_target;

	glGenTextures(1, &m_n_id);
	if(m_n_id > 0) {
		p_state->DisableTexture1D();
		p_state->DisableTexture2D();
		p_state->EnableTexture3D();
		p_state->DisableTextureCubeMap();

		p_state->BindTexture3D(m_n_id);
		
		if(b_create_mipmaps)
			glTexParameteri(m_n_target, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage3D(m_n_target, 0, m_n_internal_format, m_n_width, m_n_height, m_n_depth,
			m_n_border_width, n_format, n_data_type, (p_src_data)? p_src_data : NULL);

		if(!b_create_mipmaps) {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		} else {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
		// t_odo - see if this is necessary - seems to be ok

		m_tex_parameters.UpdateState(m_n_target);

		if(glGetError() != GL_NO_ERROR) {
			glDeleteTextures(1, &m_n_id);
			m_n_id = 0;
		}
	}
}

/*
 *	void CGLTexture_3D::Bind(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 */
void CGLTexture_3D::Bind(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_3D::Bind");

	p_state->BindTexture3D(m_n_id);
}

/*
 *	void CGLTexture_3D::Bind_Enable(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 *		- enables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_3D::Bind_Enable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_3D::Bind_Enable");

	p_state->EnableTexture3D();
	p_state->BindTexture3D(m_n_id);
}

/*
 *	void CGLTexture_3D::Disable(CGLState *p_state) const
 *		- disables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_3D::Disable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_3D::Disable");

	p_state->DisableTexture3D();
}

/*
 *	CGLTextureParams &CGLTexture_3D::r_Parameters(CGLState *p_state)
 *		- enables texture parameters to be changed
 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
 */
CGLTextureParams &CGLTexture_3D::r_Parameters(CGLState *p_state)
{
	__FuncGuard("CGLTexture_3D::r_Parameters");

	p_state->EnableTexture3D();
	p_state->BindTexture3D(m_n_id);
	return m_tex_parameters;
}

/*
 *								=== ~CGLTexture_3D ===
 */

/*
 *								=== CGLTexture_Cube ===
 */

GLenum CGLTexture_Cube::p_cube_face[6] = {
	GL_TEXTURE_CUBE_MAP_POSITIVE_X,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
};

CGLTexture_Cube::CGLTexture_Cube(CGLState *p_state, int n_width, GLenum n_internal_format,
	bool b_create_mipmaps, bool b_software_mipmaps, int n_border_width,
	GLenum n_target, GLenum n_format, GLenum n_data_type, const void *p_src_data[6])
	:m_n_width(n_width)
{
	__FuncGuard("CGLTexture_Cube::CGLTexture_Cube");

	m_n_internal_format = n_internal_format;
	m_n_border_width = n_border_width;
	m_n_target = n_target;

	glGenTextures(1, &m_n_id);
	if(m_n_id > 0) {
		p_state->DisableTexture1D();
		p_state->DisableTexture2D();
		p_state->DisableTexture3D();
		p_state->EnableTextureCubeMap();

		p_state->BindTextureCubeMap(m_n_id);
		
		if(b_create_mipmaps)
			glTexParameteri(m_n_target, GL_GENERATE_MIPMAP, GL_TRUE);

		for(int i = 0; i < 6; ++ i) {
			glTexImage2D(p_cube_face[i], 0, m_n_internal_format, m_n_width, m_n_width,
				m_n_border_width, n_format, n_data_type, (p_src_data)? p_src_data[i] : NULL);
		}

		if(!b_create_mipmaps) {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		} else {
			glTexParameteri(m_n_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			glTexParameteri(m_n_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
		// t_odo - see if this is necessary - seems to be ok

		m_tex_parameters.UpdateState(m_n_target);

		if(glGetError() != GL_NO_ERROR) {
			glDeleteTextures(1, &m_n_id);
			m_n_id = 0;
		}
	}
}

/*
 *	void CGLTexture_Cube::Bind(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 */
void CGLTexture_Cube::Bind(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_Cube::Bind");

	p_state->BindTextureCubeMap(m_n_id);
}

/*
 *	void CGLTexture_Cube::Bind_Enable(CGLState *p_state) const
 *		- binds texture object to OpenGL state as active texture in current texture unit
 *		- enables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_Cube::Bind_Enable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_Cube::Bind_Enable");

	p_state->EnableTextureCubeMap();
	p_state->BindTextureCubeMap(m_n_id);
}

/*
 *	void CGLTexture_Cube::Disable(CGLState *p_state) const
 *		- disables application of texture target (i.e. GL_TEXTURE_1D or similar)
 */
void CGLTexture_Cube::Disable(CGLState *p_state) const
{
	__FuncGuard("CGLTexture_Cube::Disable");

	p_state->DisableTextureCubeMap();
}

/*
 *	CGLTextureParams &CGLTexture_Cube::r_Parameters(CGLState *p_state)
 *		- enables texture parameters to be changed
 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
 */
CGLTextureParams &CGLTexture_Cube::r_Parameters(CGLState *p_state)
{
	__FuncGuard("CGLTexture_Cube::r_Parameters");

	p_state->EnableTextureCubeMap();
	p_state->BindTextureCubeMap(m_n_id);
	return m_tex_parameters;
}

/*
 *								=== ~CGLTexture_Cube ===
 */
